(************************************************************************
 Author:	Robert Rossmair
		Wagelwitz, Rodaer Str. 6
		04688 Mutzschen
		Germany

	e-mail:	Robert.Rossmair@t-online.de

 Module:	MDIWallp

 Version:	1.5, 06-Mar-1998

 History:
   06-Mar-98: Version 1.5
	Added transparency code by Mike Burton (see below),
	with some modifications, among others:
	- SetTransparent property access method
	- TransparentColor property
	- TransparentMode property
   #########################################
   05-Mar-1998
      Michael Burton AMS-Services, Inc.
      burtonmi@ams.iix.com

      Add a transparent background to the
      bitmap. Use the upper-left pixel of
      the bitmap as the transparent color.

      Add a Transparent property to control
      the transparent background.
   #########################################
   03-Mar-98: Version 1.4
	Bug fix: Prior versions will not redraw correctly on
	form resize if Mode property not in [wpTile, wpTopLeft].
   09-Dez-97: Version 1.3
	Make it work with non-MDI windows by commenting
	out two lines in TMDIWallpaper.Create.
	Minor fix.
   28-Nov-97: Version 1.2
	Added wpTopRight, wpBottomLeft, wpBottomRight.
	Some fixes.
   06-Nov-97: Version 1.1
	added wpStretch
   22-Oct-97: released Version 1.0

 Description:	Background bitmap component for MDI frame windows

 Copyright  1997  Robert Rossmair

 Use, modification and distribution is allowed without limitation,
 warranty, or liability of any kind.

 If you use this software or portions of it in a commercial program,
 I would appreciate it if you mentioned my contribution in the
 documentation materials accompaning your work.
************************************************************************)
unit MDIWallp;

interface

uses SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms;

type
  TWallpaperMode = (wpTile, wpCenter, wpTopLeft, wpStretch, wpTopRight, wpBottomLeft, wpBottomRight);
{$IFDEF VER90}
  TTransparentMode = (tmAuto, tmFixed);
{$ENDIF}
  EMDIWallpaper = class(Exception);

  TMDIWallpaper = class(TComponent)
  private
    FClientInstance: Pointer;
    FDefClientProc: TFarProc;
    FBitmap: TBitmap;
    FMode: TWallpaperMode;
    FTransparent: Boolean;
    FTransparentColor: TColor;
    FTransparentMode: TTransparentMode;
    procedure ClientInvalidate;
    procedure ClientWndProc(var Message: TMessage);
    procedure ConnectToClient;
    function GetClient: TForm;
    function GetClientHandle: hWnd;
    function GetTransparentColor: TColor;
    procedure SetBitmap(Value: TBitmap);
    procedure SetMode(Value: TWallpaperMode);
    procedure SetTransparent(Value: Boolean);
    procedure SetTransparentColor(Value: TColor);
    procedure SetTransparentMode(Value: TTransparentMode);
    function TransparentColorStored: Boolean;
  protected
    destructor Destroy; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Client: TForm read GetClient;
    property ClientHandle: hWnd read GetClientHandle;
  published
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property Mode: TWallpaperMode read FMode write SetMode;
    property Transparent: Boolean read FTransparent write SetTransparent;
    property TransparentColor: TColor read GetTransparentColor
      write SetTransparentColor stored TransparentColorStored;
    property TransparentMode: TTransparentMode read FTransparentMode
      write SetTransparentMode default tmAuto;
  end;

procedure Register;

implementation

procedure Error(Text: string);
var
  StackTop: record end;
  Stack: record
    EBP: Integer;
    ReturnAddress: Pointer;
  end absolute StackTop;
begin
  raise EMDIWallpaper.Create(Text) at Stack.ReturnAddress;
end;

constructor TMDIWallpaper.Create(AOwner: TComponent);
var
  i: Integer;
begin
  if not (AOwner is TForm) then
    Error('not (TMDIWallpaper.Owner is TForm)!')
{ commented out to make this work with non-MDI windows, too:
   else if TForm(AOwner).FormStyle <> fsMDIForm then
    Error('This is no MDI frame window')
}
  else with AOwner do for i := 0 to ComponentCount-1 do if Components[i] is TMDIWallpaper then
    Error('Only one instance of TMDIWallpaper allowed');
  inherited Create(AOwner);
  FTransparentColor := clDefault;
  FBitmap := TBitmap.Create;
  TForm(Owner).HandleNeeded;
  ConnectToClient;
end;

destructor TMDIWallpaper.Destroy;
begin
  if Assigned(Owner) and not (csDesigning in ComponentState) then
  begin
    if Client.HandleAllocated then
    begin
{ Falls jemand Anderer es wagt, in dreister Miachtung der
  Maximen objektorientierter Programmierung ebenfalls in
  den Innereien des Clients herumzupfuschen (WIR drfen das),
  ist hier Gelegenheit, das ans Licht zu bringen:
FREE translation:
  If anybody else dares to tamper with the Client (WE are allowed to do
  so-), we'll find it out NOW:
}
      if SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FDefClientProc)) <> Longint(FClientInstance)
      then Error('Internal Error in TMDIWallpaper.Destroy');
      FreeObjectInstance(FClientInstance);
    end
    else FreeObjectInstance(FDefClientProc);
  end;
  FBitmap.Free;
  inherited;
end;

procedure TMDIWallpaper.ConnectToClient;
{ This is the dirty way, but heck, who knows a better one? }
begin
  if csDesigning in ComponentState then Exit;
  FClientInstance := MakeObjectInstance(ClientWndProc);
  FDefClientProc := Pointer(GetWindowLong(ClientHandle, GWL_WNDPROC));
  SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FClientInstance));
end;

procedure TileBitmap(Bitmap: TBitmap; Canvas: TCanvas; Width, Height: Integer);
var
  i, j: Integer;
begin
  for i := 0 to (Width-1) div Bitmap.Width do
  for j := 0 to (Height-1) div Bitmap.Height do
    Canvas.Draw(i*Bitmap.Width, j*Bitmap.Height, Bitmap);
end;

procedure TMDIWallpaper.ClientInvalidate;
var
  Message: TWMEraseBkgnd;
begin
  if not (csDesigning in ComponentState) then
  with Message do
  begin
    Msg := wm_EraseBkgnd;
    Unused := 0;
    Result := 0;
    DC := GetDC(ClientHandle);
    try
      ClientWndProc(TMessage(Message));
    finally
      ReleaseDC(ClientHandle, DC);
    end;
  end;
end;

procedure TMDIWallpaper.ClientWndProc(var Message: TMessage);
var
  Bmp: TBitmap;
  R: TRect;

  procedure Default;
  begin
    with Message do
      Result := CallWindowProc(FDefClientProc, ClientHandle, Msg, wParam, lParam);
  end;

  procedure PaintBackground(Bitmap: TBitmap);
  var
    X, Y, W, H: Integer;
  begin
    with Client.Canvas do
    begin
      Handle := TWMEraseBkGnd(Message).DC;
      Brush := Client.Brush;
      Windows.GetClientRect(ClientHandle, R);
      W := R.Right-R.Left;
      H := R.Bottom-R.Top;
      case FMode of
	wpTile: TileBitmap(Bitmap, Client.Canvas, W, H);
	wpCenter:
		begin
		  X := (W - Bitmap.Width) div 2;
		  Y := (H - Bitmap.Height) div 2;
		  FillRect(Rect(0, 0, W, Y));
		  FillRect(Rect(0, Y, X, Y+Bitmap.Height));
		  FillRect(Rect(X+Bitmap.Width, Y, W, Y+Bitmap.Height));
		  FillRect(Rect(0, Y+Bitmap.Height, W, H));
		  Draw(X, Y, Bitmap);
		end;
	wpTopLeft:
		begin
		  FillRect(Rect(Bitmap.Width, 0, W, Bitmap.Height));
		  FillRect(Rect(0, Bitmap.Height, W, H));
		  Draw(0, 0, Bitmap);
		end;
	wpStretch:
		StretchDraw(R, Bitmap);
	wpTopRight:
		begin
		  FillRect(Rect(0, 0, W-Bitmap.Width, Bitmap.Height));
		  FillRect(Rect(0, Bitmap.Height, W, H));
		  Draw(W-Bitmap.Width, 0, Bitmap);
		end;
	wpBottomLeft:
		begin
		  FillRect(Rect(0, 0, W, H-Bitmap.Height));
		  FillRect(Rect(Bitmap.Width, H-Bitmap.Height, W, H));
		  Draw(0, H-Bitmap.Height, Bitmap);
		end;
	wpBottomRight:
		begin
		  FillRect(Rect(0, 0, W, H-Bitmap.Height));
		  FillRect(Rect(0, H-Bitmap.Height, W-Bitmap.Width, H));
		  Draw(W-Bitmap.Width, H-Bitmap.Height, Bitmap);
		end;
      end;
      Handle := 0;
    end;
  end;

begin
  with Message do
    if (Msg = WM_SIZE) and not (FMode in [wpTile, wpTopLeft]) then
    begin
      Windows.GetClientRect(ClientHandle, R);
      Windows.InvalidateRect(ClientHandle, @R, True);
    end else
    if (Msg = WM_ERASEBKGND) then
    if Bitmap.Width > 0 then
    if Bitmap.Height > 0 then
    begin
      if FTransparent then
      { code by	Michael Burton AMS-Services, Inc.
		burtonmi@ams.iix.com
	modified by RR }
      with TImageList.CreateSize(Bitmap.Width, Bitmap.Height) do
      try
	Bmp := TBitmap.Create;
	try
	  Bmp.Assign(Bitmap);
	  AddMasked(Bmp, GetTransparentColor);
{$IFNDEF VER90}
	  Bmp.HandleType := bmDDB; // otherwise eventually background color won't appear correctly
{$ENDIF}	  
	  BkColor := ColorToRGB(Client.Color);
	  Draw(Bmp.Canvas, 0, 0, 0);
	  PaintBackground(Bmp);
	finally
	  Bmp.Free;
	end;
      finally
	Free;
      end
      else PaintBackground(Bitmap);
      Result := 1;
      Exit;
    end;
  Default;
end;

function TMDIWallpaper.GetClient: TForm;
begin
  Result := TForm(Owner);
end;

function TMDIWallpaper.GetClientHandle: hWnd;
begin
  with Client do
  if ClientHandle <> 0
    then Result := ClientHandle
    else Result := Handle;
end;

function TMDIWallpaper.GetTransparentColor: TColor;
begin
  if (FTransparentColor = clDefault) and (FBitmap <> nil) then
  with FBitmap do
  begin
    if Monochrome then
      Result := clWhite
    else
      Result := Canvas.Pixels[0, Height - 1];
  end
  else Result := ColorToRGB(FTransparentColor);
  Result := Result or $02000000;
end;

procedure TMDIWallpaper.SetBitmap(Value: TBitmap);
begin
  FBitmap.Assign(Value);
  if Value <> nil then
  begin
    FTransparentColor := Value.TransparentColor;
{$IFNDEF VER90}
    FTransparentMode := Value.TransparentMode;
{$ENDIF}
  end;
  ClientInvalidate;
end;

procedure TMDIWallpaper.SetMode(Value: TWallpaperMode);
begin
  if Value <> FMode then
  begin
    FMode := Value;
    ClientInvalidate;
  end;
end;

procedure TMDIWallpaper.SetTransparent(Value: Boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent := Value;
    ClientInvalidate;
  end;
end;

procedure TMDIWallpaper.SetTransparentColor(Value: TColor);
begin
  if Value <> FTransparentColor then
  begin
    if Value = clDefault then
      FTransparentMode := tmAuto else
      FTransparentMode := tmFixed;
    FTransparentColor := Value;
    ClientInvalidate;
  end;
end;

procedure TMDIWallpaper.SetTransparentMode(Value: TTransparentMode);
begin
  if Value <> FTransparentMode then
  begin
    if Value = tmAuto then
      SetTransparentColor(clDefault) else
      SetTransparentColor(GetTransparentColor);
  end;
end;

function TMDIWallpaper.TransparentColorStored: Boolean;
begin
  Result := FTransparentMode = tmFixed;
end;

const
{$IFDEF Bilch}
  PalettePage = 'Homebrew';
{$ELSE}
  PalettePage = '3rdParty';
{$ENDIF}
procedure Register;
begin
  RegisterComponents(PalettePage, [TMDIWallpaper]);
end;

end.

